//***************************************************************************
// SCSI definitions
//***************************************************************************

//***************************************************************************
//       %%% TARGET STATUS VALUES %%%
//***************************************************************************
#define STATUS_GOOD			0x00   // Status Good
#define STATUS_CHKCOND		0x02   // Check Condition
#define STATUS_CONDMET		0x04   // Condition Met
#define STATUS_BUSY			0x08   // Busy
#define STATUS_INTERM		0x10   // Intermediate
#define STATUS_INTCDMET		0x14   // Intermediate-condition met
#define STATUS_RESCONF		0x18   // Reservation conflict
#define STATUS_COMTERM		0x22   // Command Terminated
#define STATUS_QFULL		0x28   // Queue full

//***************************************************************************
//      %%% SCSI MISCELLANEOUS EQUATES %%%
//***************************************************************************
#define MAXLUN				0x07	// Maximum Logical Unit Id
#define MAXTARG				0x0F	// Maximum Target Id
#define MAX_SCSI_LUNS		64		// Maximum Number of SCSI LUNs
#define MAX_NUM_HA			8		// Maximum Number of SCSI HA's

//---------------------------------------------------------------------------
//
//       %%% SCSI COMMAND OPCODES %%%
//
//---------------------------------------------------------------------------
//***************************************************************************
//     %%% Commands for all Device Types %%%
//***************************************************************************
#define SCSI_CHANGE_DEF		0x40   // Change Definition (Optional)
#define SCSI_COMPARE		0x39   // Compare (O)
#define SCSI_COPY			0x18   // Copy (O)
#define SCSI_COP_VERIFY		0x3A   // Copy and Verify (O)
#define SCSI_INQUIRY		0x12   // Inquiry (MANDATORY)
#define SCSI_LOG_SELECT		0x4C   // Log Select (O)
#define SCSI_LOG_SENSE		0x4D   // Log Sense (O)
#define SCSI_MODE_SEL6		0x15   // Mode Select 6-byte (Device Specific)
#define SCSI_MODE_SEL10		0x55   // Mode Select 10-byte (Device Specific)
#define SCSI_MODE_SEN6		0x1A   // Mode Sense 6-byte (Device Specific)
#define SCSI_MODE_SEN10		0x5A   // Mode Sense 10-byte (Device Specific)
#define SCSI_READ_BUFF		0x3C   // Read Buffer (O)
#define SCSI_REQ_SENSE		0x03   // Request Sense (MANDATORY)
#define SCSI_SEND_DIAG		0x1D   // Send Diagnostic (O)
#define SCSI_TST_U_RDY		0x00   // Test Unit Ready (MANDATORY)
#define SCSI_WRITE_BUFF		0x3B   // Write Buffer (O)

//***************************************************************************
//     %%% Commands Unique to Direct Access Devices %%%
//***************************************************************************
#define SCSI_COMPARE		0x39   // Compare (O)
#define SCSI_FORMAT			0x04   // Format Unit (MANDATORY)
#define SCSI_LCK_UN_CAC		0x36   // Lock Unlock Cache (O)
#define SCSI_PREFETCH		0x34   // Prefetch (O)
#define SCSI_MED_REMOVL		0x1E   // Prevent/Allow medium Removal (O)
#define SCSI_READ6			0x08   // Read 6-byte (MANDATORY)
#define SCSI_READ10			0x28   // Read 10-byte (MANDATORY)
#define SCSI_RD_CAPAC		0x25   // Read Capacity (MANDATORY)
#define SCSI_RD_DEFECT		0x37   // Read Defect Data (O)
#define SCSI_READ_LONG		0x3E   // Read Long (O)
#define SCSI_REASS_BLK		0x07   // Reassign Blocks (O)
#define SCSI_RCV_DIAG		0x1C   // Receive Diagnostic Results (O)
#define SCSI_RELEASE		0x17   // Release Unit (MANDATORY)
#define SCSI_REZERO			0x01   // Rezero Unit (O)
#define SCSI_SRCH_DAT_E		0x31   // Search Data Equal (O)
#define SCSI_SRCH_DAT_H		0x30   // Search Data High (O)
#define SCSI_SRCH_DAT_L		0x32   // Search Data Low (O)
#define SCSI_SEEK6			0x0B   // Seek 6-Byte (O)
#define SCSI_SEEK10			0x2B   // Seek 10-Byte (O)
#define SCSI_SEND_DIAG		0x1D   // Send Diagnostics (MANDATORY)
#define SCSI_SET_LIMIT		0x33   // Set Limits (O)
#define SCSI_START_STP		0x1B   // Start/Stop Unit (O)
#define SCSI_SYNC_CACHE		0x35   // Synchronize Cache (O)
#define SCSI_VERIFY			0x2F   // Verify (O)
#define SCSI_WRITE6			0x0A   // Write 6-Byte (MANDATORY)
#define SCSI_WRITE10		0x2A   // Write 10-Byte (MANDATORY)
#define SCSI_WRT_VERIFY		0x2E   // Write and Verify (O)
#define SCSI_WRITE_LONG		0x3F   // Write Long (O)
#define SCSI_WRITE_SAME		0x41   // Write Same (O)

//***************************************************************************
//   %%% Commands Unique to Sequential Access Devices %%%
//***************************************************************************
#define SCSI_ERASE			0x19   // Erase (MANDATORY)
#define SCSI_LOAD_UN		0x1B   // Load/Unload (O)
#define SCSI_LOCATE			0x2B   // Locate (O)
#define SCSI_RD_BLK_LIM		0x05   // Read Block Limits (MANDATORY)
#define SCSI_READ_POS		0x34   // Read Position (O)
#define SCSI_READ_REV		0x0F   // Read Reverse (O)
#define SCSI_REC_BF_DAT		0x14   // Recover Buffer Data (O)
#define SCSI_RESERVE		0x16   // Reserve Unit (MANDATORY)
#define SCSI_REWIND			0x01   // Rewind (MANDATORY)
#define SCSI_SPACE			0x11   // Space (MANDATORY)
#define SCSI_VERIFY_T		0x13   // Verify (Tape) (O)
#define SCSI_WRT_FILE		0x10   // Write Filemarks (MANDATORY)

//***************************************************************************
//      %%% Commands Unique to Printer Devices %%%
//***************************************************************************
#define SCSI_PRINT			0x0A   // Print (MANDATORY)
#define SCSI_SLEW_PNT		0x0B   // Slew and Print (O)
#define SCSI_STOP_PNT		0x1B   // Stop Print (O)
#define SCSI_SYNC_BUFF		0x10   // Synchronize Buffer (O)

//***************************************************************************
//     %%% Commands Unique to Processor Devices %%%
//***************************************************************************
#define SCSI_RECEIVE		0x08   // Receive (O)
#define SCSI_SEND			0x0A   // Send (O)

//***************************************************************************
//    %%% Commands Unique to Write-Once Devices %%%
//***************************************************************************
#define SCSI_MEDIUM_SCN   0x38   // Medium Scan (O)
#define SCSI_SRCHDATE10   0x31   // Search Data Equal 10-Byte (O)
#define SCSI_SRCHDATE12   0xB1   // Search Data Equal 12-Byte (O)
#define SCSI_SRCHDATH10   0x30   // Search Data High 10-Byte (O)
#define SCSI_SRCHDATH12   0xB0   // Search Data High 12-Byte (O)
#define SCSI_SRCHDATL10   0x32   // Search Data Low 10-Byte (O)
#define SCSI_SRCHDATL12   0xB2   // Search Data Low 12-Byte (O)
#define SCSI_SET_LIM_10   0x33   // Set Limits 10-Byte (O)
#define SCSI_SET_LIM_12   0xB3   // Set Limits 10-Byte (O)
#define SCSI_VERIFY10    0x2F   // Verify 10-Byte (O)
#define SCSI_VERIFY12    0xAF   // Verify 12-Byte (O)
#define SCSI_WRITE12    0xAA   // Write 12-Byte (O)
#define SCSI_WRT_VER10    0x2E   // Write and Verify 10-Byte (O)
#define SCSI_WRT_VER12    0xAE   // Write and Verify 12-Byte (O)

//***************************************************************************
//      %%% Commands Unique to CD-ROM Devices %%%
//***************************************************************************
#define SCSI_PLAYAUD_10   0x45   // Play Audio 10-Byte (O)
#define SCSI_PLAYAUD_12   0xA5   // Play Audio 12-Byte 12-Byte (O)
#define SCSI_PLAYAUDMSF   0x47   // Play Audio MSF (O)
#define SCSI_PLAYA_TKIN   0x48   // Play Audio Track/Index (O)
#define SCSI_PLYTKREL10   0x49   // Play Track Relative 10-Byte (O)
#define SCSI_PLYTKREL12   0xA9   // Play Track Relative 12-Byte (O)
#define SCSI_READCDCAP    0x25   // Read CD-ROM Capacity (MANDATORY)
#define SCSI_READHEADER   0x44   // Read Header (O)
#define SCSI_SUBCHANNEL   0x42   // Read Subchannel (O)
#define SCSI_READ_TOC    0x43   // Read TOC (O)

//***************************************************************************
//      %%% Commands Unique to Scanner Devices %%%
//***************************************************************************
#define SCSI_GETDBSTAT    0x34   // Get Data Buffer Status (O)
#define SCSI_GETWINDOW    0x25   // Get Window (O)
#define SCSI_OBJECTPOS    0x31   // Object Postion (O)
#define SCSI_SCAN     0x1B   // Scan (O)
#define SCSI_SETWINDOW    0x24   // Set Window (MANDATORY)

//***************************************************************************
//    %%% Commands Unique to Optical Memory Devices %%%
//***************************************************************************
#define SCSI_UpdateBlk    0x3D   // Update Block (O)

//***************************************************************************
//    %%% Commands Unique to Medium Changer Devices %%%
//***************************************************************************
#define SCSI_EXCHMEDIUM   0xA6   // Exchange Medium (O)
#define SCSI_INITELSTAT   0x07   // Initialize Element Status (O)
#define SCSI_POSTOELEM    0x2B   // Position to Element (O)
#define SCSI_REQ_VE_ADD   0xB5   // Request Volume Element Address (O)
#define SCSI_SENDVOLTAG   0xB6   // Send Volume Tag (O)

//***************************************************************************
//     %%% Commands Unique to Communication Devices %%%
//***************************************************************************
#define SCSI_GET_MSG_6    0x08   // Get Message 6-Byte (MANDATORY)
#define SCSI_GET_MSG_10   0x28   // Get Message 10-Byte (O)
#define SCSI_GET_MSG_12   0xA8   // Get Message 12-Byte (O)
#define SCSI_SND_MSG_6    0x0A   // Send Message 6-Byte (MANDATORY)
#define SCSI_SND_MSG_10   0x2A   // Send Message 10-Byte (O)
#define SCSI_SND_MSG_12   0xAA   // Send Message 12-Byte (O)
//---------------------------------------------------------------------------
//
//       %%% END OF SCSI COMMAND OPCODES %%%
//
//---------------------------------------------------------------------------

//***************************************************************************
//      %%% Request Sense Data Format %%%
//***************************************************************************
typedef struct 
{
	BYTE  ErrorCode;    // Error Code (70H or 71H)
	BYTE  SegmentNum;    // Number of current segment descriptor
	BYTE  SenseKey;    // Sense Key(See bit definitions too)
	BYTE  InfoByte0;    // Information MSB
	BYTE  InfoByte1;    // Information MID
	BYTE  InfoByte2;    // Information MID
	BYTE  InfoByte3;    // Information LSB
	BYTE  AddSenLen;    // Additional Sense Length
	BYTE  ComSpecInf0;   // Command Specific Information MSB
	BYTE  ComSpecInf1;   // Command Specific Information MID
	BYTE  ComSpecInf2;   // Command Specific Information MID
	BYTE  ComSpecInf3;   // Command Specific Information LSB
	BYTE  AddSenseCode;   // Additional Sense Code
	BYTE  AddSenQual;    // Additional Sense Code Qualifier
	BYTE  FieldRepUCode;   // Field Replaceable Unit Code
	BYTE  SenKeySpec15;   // Sense Key Specific 15th byte
	BYTE  SenKeySpec16;   // Sense Key Specific 16th byte
	BYTE  SenKeySpec17;   // Sense Key Specific 17th byte
	BYTE  AddSenseBytes;   // Additional Sense Bytes
}SENSE_DATA_FMT;

//***************************************************************************
//       %%% REQUEST SENSE ERROR CODE %%%
//***************************************************************************
#define SERROR_CURRENT    0x70   // Current Errors
#define SERROR_DEFERED    0x71   // Deferred Errors

//***************************************************************************
//      %%% REQUEST SENSE BIT DEFINITIONS %%%
//***************************************************************************
#define SENSE_VALID    0x80   // Byte 0 Bit 7
#define SENSE_FILEMRK    0x80   // Byte 2 Bit 7
#define SENSE_EOM     0x40   // Byte 2 Bit 6
#define SENSE_ILI     0x20   // Byte 2 Bit 5

//***************************************************************************
//     %%% REQUEST SENSE SENSE KEY DEFINITIONS %%%
//***************************************************************************
#define KEY_NOSENSE    0x00   // No Sense
#define KEY_RECERROR    0x01   // Recovered Error
#define KEY_NOTREADY    0x02   // Not Ready
#define KEY_MEDIUMERR    0x03   // Medium Error
#define KEY_HARDERROR    0x04   // Hardware Error
#define KEY_ILLGLREQ    0x05   // Illegal Request
#define KEY_UNITATT    0x06   // Unit Attention
#define KEY_DATAPROT    0x07   // Data Protect
#define KEY_BLANKCHK    0x08   // Blank Check
#define KEY_VENDSPEC    0x09   // Vendor Specific
#define KEY_COPYABORT    0x0A   // Copy Abort
#define KEY_EQUAL     0x0C   // Equal (Search)
#define KEY_VOLOVRFLW    0x0D   // Volume Overflow
#define KEY_MISCOMP    0x0E   // Miscompare (Search)
#define KEY_RESERVED    0x0F   // Reserved

//***************************************************************************
//      %%% PERIPHERAL DEVICE TYPE DEFINITIONS %%%
//***************************************************************************
#define DTYPE_DASD		0x00   // Disk Device
#define DTYPE_SEQD		0x01   // Tape Device
#define DTYPE_PRNT		0x02   // Printer
#define DTYPE_PROC		0x03   // Processor
#define DTYPE_WORM		0x04   // Write-once read-multiple
#define DTYPE_CROM		0x05   // CD-ROM device
#define DTYPE_SCAN		0x06   // Scanner device
#define DTYPE_OPTI		0x07   // Optical memory device
#define DTYPE_JUKE		0x08   // Medium Changer device
#define DTYPE_COMM		0x09   // Communications device
#define DTYPE_RESL		0x0A   // Reserved (low)
#define DTYPE_RESH		0x1E   // Reserved (high)
#define DTYPE_UNKNOWN	0x1F   // Unknown or no device type

//***************************************************************************
//      %%% ANSI APPROVED VERSION DEFINITIONS %%%
//***************************************************************************
#define ANSI_MAYBE     0x0  // Device may or may not be ANSI approved stand
#define ANSI_SCSI1     0x1  // Device complies to ANSI X3.131-1986 (SCSI-1)
#define ANSI_SCSI2     0x2  // Device complies to SCSI-2
#define ANSI_RESLO     0x3  // Reserved (low)
#define ANSI_RESHI     0x7  // Reserved (high)


//***************************************************************************
//      %%% ADDITIONAL SENSE CODE & ADDITIONAL SENSE CODE QUALIFIER %%%
//***************************************************************************
#define ASC_NOINFO		0x00	//No information available
#define ASQ_NOINFO		0x00	//
#define ASC_FILMRK		0x00	//Filemark detected
#define ASQ_FILMRK		0x01	//
#define ASC_EOM			0x00	//End Of Medium
#define ASQ_EOM			0x02	//
#define ASC_SETMRK		0x00	//Setmark detected
#define ASQ_SETMRK		0x03	//
#define ASC_BOM			0x00	//Begin Of Medium
#define ASQ_BOM			0x04	//
#define ASC_EOD			0x00	//End Of Data
#define ASQ_EOD			0x05	//
#define ASC_IOTERM		0x00	//I/O-Process Terminated
#define ASQ_IOTERM		0x06	//
#define ASC_PWRTFLT		0x03	//Peripheral Write Fault
#define ASQ_PWRTFLT		0x00	//
#define ASC_NOWRT		0x03	//No Write Current
#define ASQ_NOWRT		0x01	//
#define ASC_EWRTERR		0x03	//Excessive Write Errors
#define ASQ_EWRTERR		0x02	//
#define ASC_LUNNRDNC	0x04	//LUN not Ready, cause not reportable
#define ASQ_LUNNRDNC	0x00	//
#define ASC_LUNBRD		0x04	//LUN is becoming Ready
#define ASQ_LUNBRD		0x01	//
#define ASC_LUNNRDIR	0x04	//LUN not Ready, Init required
#define ASQ_LUNNRDIR	0x02	//
#define ASC_LUNNRDMI	0x04	//LUN not Ready, manual Intervention required
#define ASQ_LUNNRDMI	0x03	//
#define ASC_LUNNRDFP	0x04	//LUN not Ready, Format in progress
#define ASQ_LUNNRDFP	0x04	//
#define ASC_LUNNSEL		0x05	//LUN does not respond to Select
#define ASQ_LUNNSEL		0x00	//
#define ASC_MULTSEL		0x07	//Multiple Peripheral Devices selected
#define ASQ_MULTSEL		0x00	//
#define ASC_LUNCFAIL	0x08	//LUN Communication Failure
#define ASQ_LUNCFAIL	0x00	//
#define ASC_LUNCTIME	0x08	//LUN Communication Timeout
#define ASQ_LUNCTIME	0x01	//
#define ASC_LUNCPAR		0x08	//LUN Communication Parity Error
#define ASQ_LUNCPAR		0x02	//
#define ASC_TRKFOLOW	0x09	//Track following Error
#define ASQ_TRKFOLOW	0x00	//
#define ASC_LOGOVER		0x0A	//Error Log Overflow
#define ASQ_LOGOVER		0x00	//
#define ASC_WRITE		0x0C	//Write Error
#define ASQ_WRITE		0x00	//
#define ASC_URREAD		0x11	//Unrecovered Read Error
#define ASQ_URREAD		0x00	//
#define ASC_RDRETRY		0x11	//Read Retries Exhausted
#define ASQ_RDRETRY		0x01	//
#define ASC_TOOLONG		0x11	//Error too long to correct
#define ASQ_TOOLONG		0x02	//
#define ASC_MULREAD		0x11	//Multiple Read Errors
#define ASQ_MULREAD		0x03	//
#define ASC_IBLKRD		0x11	//Incomplete Block Read
#define ASQ_IBLKRD		0x08	//
#define ASC_NOGAP		0x11	//No Gap Found
#define ASQ_NOGAP		0x09	//
#define ASC_MISCORR		0x11	//Miscorrected Error
#define ASQ_MISCORR		0x0A	//
#define ASC_RECENTI		0x14	//Recorded Entity not found
#define ASQ_RECENTI		0x00	//
#define ASC_RECORD		0x14	//Record not found
#define ASQ_RECORD		0x01	//
#define ASC_FILESET		0x14	//File- or SetMark not found
#define ASQ_FILESET		0x02	//
#define ASC_NOEOD		0x14	//End Of Data not found
#define ASQ_NOEOD		0x03	//
#define ASC_BLKSEQ		0x14	//Block Sequence Error
#define ASQ_BLKSEQ		0x04	//
#define ASC_RNDPOS		0x15	//Random Positioning Error
#define ASQ_RNDPOS		0x00	//
#define ASC_MECHPOS		0x15	//Mechanical Positioning Error
#define ASQ_MECHPOS		0x01	//
#define ASC_RDPOSE		0x15	//Positioning Error detected by Read
#define ASQ_RDPOSE		0x02	//
#define ASC_RECDNEC		0x17	//Recovered Data With No Error Correction Applied
#define ASQ_RECDNEC		0x00	//
#define ASC_RECDRET		0x17	//Recovered Data With Retries
#define ASQ_RECDRET		0x01	//
#define ASC_RECDPHO		0x17	//Recovered Data With Positive Head Offset
#define ASQ_RECDPHO		0x02	//
#define ASC_RECDNHO		0x17	//Recovered Data With Negative Head Offset
#define ASQ_RECDNHO		0x03	//
#define ASC_RECDWEC		0x18	//Recovered Data With Error Correction Applied
#define ASQ_RECDWEC		0x00	//
#define ASC_PARLLEN		0x1A	//Parameter List Length Error
#define ASQ_PARLLEN		0x00	//
#define ASC_SYNCERR		0x1B	//Synchronous Data Transfer Error
#define ASQ_SYNCERR		0x00	//
#define ASC_INVCOC		0x20	//Invalid Command Operation Code
#define ASQ_INVCOC		0x00	//
#define ASC_LBARNG		0x21	//Logical Block Address Out Of Range
#define ASQ_LBARNG		0x00	//
#define ASC_INVFCDB		0x24	//Invalid Field In Cdb
#define ASQ_INVFCDB		0x00	//
#define ASC_NOLOGUN		0x25	//Logical Unit Not Supported
#define ASQ_NOLOGUN		0x00	//
#define ASC_INVFPL		0x26	//Invalid Field In Parameter List
#define ASQ_INVFPL		0x00	//
#define ASC_PARNSUP		0x26	//Parameter Not Supported
#define ASQ_PARNSUP		0x01	//
#define ASC_PARVINV		0x26	//Parameter Value Invalid
#define ASQ_PARVINV		0x02	//
#define ASC_NTHRPAR		0x26	//Threshold Parameters Not Supported
#define ASQ_NTHRPAR		0x03	//
#define ASC_WRTPROT		0x27	//Write Protected
#define ASQ_WRTPROT		0x00	//
#define ASC_NRDYRDY		0x28	//Not Ready To Ready Transition(Medium May Have Changed)
#define ASQ_NRDYRDY		0x00	//
#define ASC_POWRON		0x29	//Power On, Reset, Or Bus Device Reset Occurred
#define ASQ_POWRON		0x00	//
#define ASC_PARCHN		0x2A	//Parameters Changed
#define ASQ_PARCHN		0x00	//
#define ASC_MPARCHN		0x2A	//Mode Parameters Changed
#define ASQ_MPARCHN		0x01	//
#define ASC_LPARCHN		0x2A	//Log Parameters Changed
#define ASQ_LPARCHN		0x02	//
#define ASC_NCPYDIS		0x2B	//Copy Cannot Execute Since Host Cannot Disconnect
#define ASQ_NCPYDIS		0x00	//
#define ASC_CMDSEQ		0x2C	//Command Sequence Error
#define ASQ_CMDSEQ		0x00	//
#define ASC_CMDCLR		0x2F	//Commands Cleared By Another Initiator
#define ASQ_CMDCLR		0x00	//
#define ASC_INCPMED		0x30	//Incompatible Medium Installed
#define ASQ_INCPMED		0x00	//
#define ASC_CANTRDU		0x30	//Cannot Read Medium - Unknown Format
#define ASQ_CANTRDU		0x01	//
#define ASC_CANTRDI		0x30	//Cannot Read Medium - Incompatible Format
#define ASQ_CANTRDI		0x02	//
#define ASC_CLNCARD		0x30	//Cleaning Cartridge Installed
#define ASQ_CLNCARD		0x03	//
#define ASC_MFRMCRP		0x31	//Medium Format Corrupted
#define ASQ_MFRMCRP		0x00	//
#define ASC_TAPELEN		0x33	//Tape Length Error
#define ASQ_TAPELEN		0x00	//
#define ASC_RNDPAR		0x37	//Rounded Parameter
#define ASQ_RNDPAR		0x00	//
#define ASC_SAVPAR		0x39	//Saving Parameters Not Supported
#define ASQ_SAVPAR		0x00	//
#define ASC_NOTAPE		0x3A	//No Tape In Drive
#define ASQ_NOTAPE		0x00	//
#define ASC_SEQPOS		0x3B	//Sequential Positioning Error
#define ASQ_SEQPOS		0x00	//
#define ASC_POSBOM		0x3B	//Tape Position Error At Beginning-of-medium
#define ASQ_POSBOM		0x01	//
#define ASC_POSEOM		0x3B	//Tape Position Error At End-of-medium
#define ASQ_POSEOM		0x02	//
#define ASC_REPOS		0x3B	//Reposition Error
#define ASQ_REPOS		0x08	//
#define ASC_INVBITS		0x3D	//Invalid Bits In Identify Message
#define ASQ_INVBITS		0x00	//
#define ASC_LUNSCFG		0x3E	//Logical Unit Has Not Self-configured Yet
#define ASQ_LUNSCFG		0x00	//
#define ASC_TRGOCHG		0x3F	//Target Operating Conditions Have Changed
#define ASQ_TRGOCHG		0x00	//
#define ASC_CHGUC		0x3F	//Microcode Has Been Changed
#define ASQ_CHGUC		0x01	//
#define ASC_CHGOD		0x3F	//Changed Operating Definition
#define ASQ_CHGOD		0x02	//
#define ASC_CHGINQ		0x3F	//Inquiry Data Has Changed
#define ASQ_CHGINQ		0x03	//
#define ASC_DGFAIL		0x40	//Diagnostic Failure On Component Nn (80h-ffh)
#define ASQ_DGFAIL		0x80	//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#define ASC_MSGERR		0x43	//Message Error
#define ASQ_MSGERR		0x00	//
#define ASC_INTTRGF		0x44	//Internal Target Failure
#define ASQ_INTTRGF		0x00	//
#define ASC_SELRESF		0x45	//Select Or Reselect Failure
#define ASQ_SELRESF		0x00	//
#define ASC_SFTRSTF		0x46	//Unsuccessful Soft Reset
#define ASQ_SFTRSTF		0x00	//
#define ASC_PARITY		0x47	//Scsi Parity Error
#define ASQ_PARITY		0x00	//
#define ASC_INDEMSG		0x48	//Initiator Detected Error Message Received
#define ASQ_INDEMSG		0x00	//
#define ASC_INVMSGE		0x49	//Invalid Message Error
#define ASQ_INVMSGE		0x00	//
#define ASC_CMDPHSE		0x4A	//Command Phase Error
#define ASQ_CMDPHSE		0x00	//
#define ASC_DATPHSE		0x4B	//Data Phase Error
#define ASQ_DATPHSE		0x00	//
#define ASC_LGUFSLF		0x4C	//Logical Unit Failed Self-configuration
#define ASQ_LGUFSLF		0x00	//
#define ASC_OVRCMD		0x4E	//Overlapped Commands Attempted
#define ASQ_OVRCMD		0x00	//
#define ASC_WRTAPE		0x50	//Write Append Error
#define ASQ_WRTAPE		0x00	//
#define ASC_WRTAPSE		0x50	//Write Append Position Error
#define ASQ_WRTAPSE		0x01	//
#define ASC_POSETIM		0x50	//Position Error Related To Timing
#define ASQ_POSETIM		0x02	//
#define ASC_ERSFAIL		0x51	//Erase Failure
#define ASQ_ERSFAIL		0x00	//
#define ASC_CRDFAIL		0x52	//Cartridge Fault
#define ASQ_CRDFAIL		0x00	//
#define ASC_MDLFAIL		0x53	//Media Load Or Eject Failed
#define ASQ_MDLFAIL		0x00	//
#define ASC_UNTAPF		0x53	//Unload Tape Failure
#define ASQ_UNTAPF		0x01	//
#define ASC_MDRMPRV		0x53	//Medium Removal Prevented
#define ASQ_MDRMPRV		0x02	//
#define ASC_OPERREQ		0x5A	//Operator Request Or State Change Input (Unspecified)
#define ASQ_OPERREQ		0x00	//
#define ASC_OPERMRR		0x5A	//Operator Medium Removal Request
#define ASQ_OPERMRR		0x01	//
#define ASC_OPERSWP		0x5A	//Operator Selected Write Protect
#define ASQ_OPERSWP		0x02	//
#define ASC_OPERSWE		0x5A	//Operator Selected Write Permit
#define ASQ_OPERSWE		0x03	//
#define ASC_LOGTRAP		0x5B	//Log Exception
#define ASQ_LOGTRAP		0x00	//
#define ASC_THRCND		0x5B	//Threshold Condition Met
#define ASQ_THRCND		0x01	//
#define ASC_LGCNTMX		0x5B	//Log Counter At Maximum
#define ASQ_LGCNTMX		0x02	//
#define ASC_LGLSCDE		0x5B	//Log List Codes Exhausted
#define ASQ_LGLSCDE		0x03	//
